//////////////////////////////////////////////////////////////////////////
// Code copyright 2005 Dave Hettmer
// Dave Hettmer copyright 1962 James Hettmer & Jane Littlefield
// Anyone taking the above copyrights seriously is copyrighted by Idiots 'R Us (insert year of birth)
//
// The primary purpose of this code is to ensure that the wings and arms maintain the
// desired animation.  UT don't know nothin' 'bout flappin' no wings, so the blending
// that happens at the waist for firing replaces the wings as well as the arms and head.
// For walking animation this isn't much on an issue, but it completely trashes the
// flying animation because not only do the wings stop flapping, but the gun and head
// are pointed at the ground because the gargoyle is horizontal while flying.
//
// Running firing requires blending at the shoulders only, and *in theory* vertical can
// happen on the spine as normal.  Except there's those wings again.  Even with walking
// and crouching, the wings wobble a bit as the gargoyle steps, and the total replacement
// from the waist up trashes that.  That would be kind of acceptable, except for jumping.
// As far as I can tell, UT treats jumping like walking and blends the same firing
// animation for mode modes.  But the problem there is that jumping uses wings, so
// UT's work to graft firing onto the model while jumping trashes the wings.  And a
// winged creature just oughta use its wings while jumping, ya know?
//
// So it simply isn't enough to manage only the running animation.  I need to manage the
// vertical animation, too.
//
// I haven't found a mao that allows swimming, so I'm not working too hard on that.
// Crouching, as far as I can tell, can be left to UT.
//
// That's a lot of churning to do, and I hear that the "clean" way to deal with that is
// to create a mod that allows me to create a player controller.  I also hear that
// that's a lot of hassle.  For a lot of reasons (first UT model and therefore there's a
// limit to how much effort is warrented, lots of real life responsibilities, no real
// connections in the UT community to get help from, don't know much about the engine, etc.),
// I did this with a parasite class that monitors the gargoyle.  This class is attached
// to the gargoyle and presumably runs Tick()s independently of the model.  The code
// assumes UT has set the gargoyle to do something bipedal, and that it may be necessary
// to quickly change that setup.
//
// Note that because this code assumes that it asynchronously from the gargoyle's code
// it can't rely on xPawn variables like bIsWalking, bIsCrouched, or bWasCrouched, and
// so on.  I expect that this code will miss state transitions in the gargoyle
// if it relies on the gargoyle's state variables.
//////////////////////////////////////////////////////////////////////////

class AnimWatcher extends Actor;

var Name PrevAnimsByChannel[16];
var Name WantedAnimsByChannel[16];

var xPawn Gargoyle;
var string GargoyleName;

var Name CurrAnim;
var Name PrevAnim;
var Name CurrAnim1;
var Name PrevAnim1;

var Name PrevSeqName;

var int TickCount;
var bool TweakedWaistRotation;
var bool TweakedDualAssault;

var bool IsVertical;

var bool WasFiring;
var bool IsFiring;
var bool ChangedFiring;

var bool WasRunning;
var bool IsRunning;
var bool ChangedRunning;

var bool WasCrouching;
var bool IsCrouching;
var bool ChangedCrouching;

var int LArmChannel;
var int RArmChannel;
var Name LUpperArmBone;
var Name RUpperArmBone;

// These are here not because they must persist, but
// because I suspect putting lots of variables on the
// stack when a function is called is a performance
// problem.

function SetupWatcher(xPawn P)
{
  local int count;
  TickCount = 0;
  for (count = 0; count < 16; count++)
    {
    PrevAnimsByChannel[count] = 'None';
    WantedAnimsByChannel[count] = '';
    }

  Gargoyle = P;
  if (Gargoyle.PlayerReplicationInfo != none)
    GargoyleName = Gargoyle.PlayerReplicationInfo.PlayerName;

  WasFiring = false;
  IsFiring = false;
  ChangedFiring = false;

  WasRunning = false;
  IsRunning = false;
  ChangedRunning = false;

  WasCrouching = false;
  IsCrouching = false;
  ChangedCrouching = false;

  IsVertical = false;

  TweakedWaistRotation = false;
  TweakedDualAssault = false;

  // Comment from the following page: http://unreal.student.utwente.nl/uncodex-ut2003/Source_engine/pawn.html#227
  // "Note that animation channels 2 through 11 are used for animation updating"
  // Dunno what exactly the scope of that is, so I'll play it safe and use
  // these channels.  I've never seen UT use a channel above 1, though.
  LArmChannel = 12;
  RArmChannel = 13;
  LUpperArmBone = 'Bip01 L UpperArm';
  RUpperArmBone = 'Bip01 R UpperArm';

  CurrAnim = 'None';
  PrevAnim = 'None';
  CurrAnim1 = 'None';
  PrevAnim1 = 'None';
}

defaultproperties
{
  bHidden = true
}

function CorrectDualAssault()
{
  local AssaultRifle Rifle;
  local rotator R;
//  local WeaponAttachment WA;
  local bool CurrentlyHoldingDualAssault;

  CurrentlyHoldingDualAssault = false;
  R.pitch = 0;
  R.yaw = 0;
  R.roll = 16000;

//  if ((Gargoyle.Weapon != none) && (Gargoyle.Weapon.ThirdPersonActor != none))
//    {
    if (Gargoyle.Weapon.IsA('AssaultRifle'))
      {
      Rifle = AssaultRifle(Gargoyle.Weapon);
      if (Rifle.OffhandActor != none)
        {
        CurrentlyHoldingDualAssault = true;
        if (!TweakedDualAssault)
          {
log("Tweaking L dual assault");
          // ...and we haven't tweaked the hand offset yet
          Gargoyle.SetBoneRotation('Bip01 L Hand', R,, 1.0);
          TweakedDualAssault = true;
          }
        }
      }
 //   }

  if (!CurrentlyHoldingDualAssault && TweakedDualAssault)
    {
log("Untweaking L dual assault");
    Gargoyle.SetBoneRotation('Bip01 L Hand',,, 0.0);
    TweakedDualAssault = false;
    }
}

function UntweakWaistRotation()
{
//log("+++ UntweakWaistRotation");
  if (!TweakedWaistRotation)
    return;
  TweakedWaistRotation = false;
  Gargoyle.SetBoneRotation(Gargoyle.SpineBone1,,, 0.0);
}

function LogCurrentAnimation()
{
  local float Time;
  local float Rate;
  local int Channel;
  local Name SeqName;

  for (Channel = 0; Channel < 16; Channel++)
    {
    Gargoyle.GetAnimParams(Channel, SeqName, Time, Rate);
    log("    Channel " $ Channel $ " is running " $ SeqName);
    }
}

function SynchronizeTorsoAnimation()
{
  local float Time;
  local float Rate;
  local Name SeqName;
  local Name CurrSeqName1;

  // Initiate the flapping (or whatever the wings are doing) at the same
  // frame as channel 0.
  // This should reduce the amount of stutter to the point where we only see
  // something wrong if the gargoyle has already rendered a frame with the
  // buggered channel 1 info before this code gets a chance to run.
//LogCurrentAnimation();
  Gargoyle.GetAnimParams(1, CurrSeqName1, Time, Rate);
  Gargoyle.GetAnimParams(0, SeqName, Time, Rate);
  if (CurrSeqName1 == SeqName)
    return;
//log("  -->SynchronizeTorsoAnimation current:" $ CurrSeqName1 $ " want:" $ SeqName);
  Gargoyle.AnimBlendParams(1, 1.0,,, Gargoyle.SpineBone1);
  Gargoyle.SetAnimFrame(Time / Rate, 1, 0);
  Gargoyle.LoopAnim(SeqName,,, 1);
  RecordWantedAnim(1, SeqName);
}

function StartFiringWhileRunning(string Direction)
{
  local Name FiringSequence;
//log("+++ StartFiringWhileRunning");

  // There must be a way to build a Name from components,
  // but I haven't found it yet.
  if (Direction == "F")
    FiringSequence = 'hRifle_BurstF';
  else if (Direction == "B")
    FiringSequence = 'hRifle_BurstB';
  else if (Direction == "R")
    FiringSequence = 'hRifle_BurstR';
  else if (Direction == "L")
    FiringSequence = 'hRifle_BurstL';

  // UT tried to start firing while running, but failed because there is no
  // firing animation named as UT is expecting.
  SynchronizeTorsoAnimation();

  // Start up the correct firing animation.
  Gargoyle.AnimBlendParams(LArmChannel, 1.0,,, LUpperArmBone);
  Gargoyle.AnimBlendParams(RArmChannel, 1.0,,, RUpperArmBone);
  Gargoyle.LoopAnim(FiringSequence,,, LArmChannel);
  Gargoyle.LoopAnim(FiringSequence,,, RArmChannel);
  RecordWantedAnim(LArmChannel, FiringSequence);
  RecordWantedAnim(RArmChannel, FiringSequence);
}

function StopFiringWhileRunning()
{
//log("+++ StopFiringWhileRunning");
  Gargoyle.AnimBlendParams(LArmChannel, 0.0,,, LUpperArmBone);
  Gargoyle.AnimBlendParams(RArmChannel, 0.0,,, RUpperArmBone);

  if (IsRunning)
    {
    // UT will put the Idle animation in place of the firing animation.
    SynchronizeTorsoAnimation();
    }
}

function StartFiringWhileVertical()
{
  local rotator R;
  local Name FiringSequence;

  FiringSequence = Gargoyle.GetAnimSequence();
//log("+++ StartFiringWhileVertical " $ FiringSequence $ " IsVertical:" $ IsVertical $ " bIsWalking:" $ Gargoyle.bIsWalking $ " bIsIdle:" $ Gargoyle.bIsIdle);

  // The hips are straight on the walk, but the firing animation was made at an angle.
  // That means the body animation that's been blended onto the waist has
  // an inappropriate twist.
  // Except: jumping and standing animation was make at an angle, so we don't need to
  // tweak that.
  FiringSequence = Gargoyle.GetAnimSequence();
  if (Left(FiringSequence, 3) == "Wal")
    FiringSequence = 'wRifle_Burst';
  else
    FiringSequence = 'vRifle_Burst';

  if (Gargoyle.bIsWalking && !Gargoyle.bIsIdle)
    {
//log("    tweaked waist rotation");
    // Since I made the model where Y is up, not Z, I gotta fix this through roll, not yaw.
    R.pitch = 0;
    R.yaw = 0;
    R.roll = 1250;
//    Gargoyle.SetBoneRotation(Gargoyle.SpineBone1, R,, 1.0);
//    TweakedWaistRotation = true;
    }
	// Load up the firing sequence.  Use the same channel UT would.
  Gargoyle.AnimBlendParams(1, 1.0,,, Gargoyle.SpineBone1);
  Gargoyle.LoopAnim(FiringSequence,,, 1);
  RecordWantedAnim(1, FiringSequence);
}

function StopFiringWhileVertical()
{
//log("+++ StopFiringWhileCrouching");
  Gargoyle.AnimBlendParams(1, 0.0,,, Gargoyle.SpineBone1);
}

function StartFiringWhileCrouching()
{
  local Name FiringSequence;
  FiringSequence = 'cCrouch_Burst';

  // Make sure we get the right stuff happening above the waist.
  // Initial testing of going directly to crouch and fire when the
  // starting a match caused UT to put Weapon_Switch on channel 1
  // and that screwed up the waist and above.
  SynchronizeTorsoAnimation();

//log("+++ StartFiringWhileCrouching");
  Gargoyle.AnimBlendParams(LArmChannel, 1.0,,, LUpperArmBone);
  Gargoyle.AnimBlendParams(RArmChannel, 1.0,,, RUpperArmBone);
  Gargoyle.LoopAnim(FiringSequence,,, LArmChannel);
  Gargoyle.LoopAnim(FiringSequence,,, RArmChannel);
  RecordWantedAnim(LArmChannel, FiringSequence);
  RecordWantedAnim(RArmChannel, FiringSequence);
}

function StopFiringWhileCrouching()
{
//log("+++ StopFiringWhileCrouching");
  Gargoyle.AnimBlendParams(LArmChannel, 0.0,,, LUpperArmBone);
  Gargoyle.AnimBlendParams(RArmChannel, 0.0,,, RUpperArmBone);
  if (IsCrouching)
    {
    // UT will put the Idle animation in place of the firing animation.
    SynchronizeTorsoAnimation();
    }
}

function ReportWantedAnim()
{
  local int channel;
  local float Time;
  local float Rate;
  local Name SeqName;

  SeqName = Gargoyle.GetAnimSequence();
  log("~~~WantedAnimsByChannel  currseq:" $ SeqName);
  for (channel = 0; channel < 16; channel++)
    {
    Gargoyle.GetAnimParams(channel, SeqName, Time, Rate);
    log("    Want " $ channel $ ":" $ WantedAnimsByChannel[channel] $ " Have:" $ SeqName $ " Time:" $ Time $ " Rate:" $ Rate);
    }
}
function RecordWantedAnim(int channel, Name SeqName)
{
//if (channel == 1) log("-->Want " $ channel $ ":" $ SeqName);
  WantedAnimsByChannel[channel] = SeqName;
}

function PlayWeaponSwitch()
{
//log("-->WeaponSwitch");
  RecordWantedAnim(LArmChannel, 'WeaponSwitch');
  RecordWantedAnim(RArmChannel, 'WeaponSwitch');
  Gargoyle.AnimBlendParams(LArmChannel, 1.0,,, LUpperArmBone);
  Gargoyle.AnimBlendParams(RArmChannel, 1.0,,, RUpperArmBone);
  Gargoyle.PlayAnim('Weapon_Switch',,, LArmChannel);
  Gargoyle.PlayAnim('Weapon_Switch',,, RArmChannel);
}

function InvokeWantedAnim()
{
  local int channel;
  local float Time;
  local float Rate;
  local Name CurrSeq;
  local bool DoingWhatWeWant;

//ReportWantedAnim();
  for (channel = 0; channel < 16; channel++)
    {
    if (WantedAnimsByChannel[channel] == '')
      continue; // Nothing desired on this channel.
    Gargoyle.GetAnimParams(channel, CurrSeq, Time, Rate);
    DoingWhatWeWant = CurrSeq == WantedAnimsByChannel[channel];
    if (DoingWhatWeWant)
      {
      // It's doing what we want.  Stop watching it for non-run sequences.
      // However, for run and crouch animations we need to monitor the channel
      // for a while because UT will do things like start Idle_Rifle after
      // we're done firing that bugger up the torso animation again.
      // Idle_Rifle while walking is OK.
      // Unfortunately, UT *also* does things we want, like Weapon_Switch.
      // Those we handle elsewhere.
      if (("Run" != Left(CurrSeq, 3)) && ("Cro" != Left(CurrSeq, 3)))
        WantedAnimsByChannel[channel] = ''; // Stop watching non-run sequences
      continue;
      }
    // UT isn't doing what we want.  However, in the case of weapon switch,
    // what we want may be wrong.
    if (CurrSeq == 'Weapon_Switch')
      PlayWeaponSwitch();
    else
      {
//log("-->Invoking " $ channel $ ":" $ WantedAnimsByChannel[channel]);
      Gargoyle.LoopAnim(WantedAnimsByChannel[channel],,, channel);
      }
    }
}

function ReportChangedAnimation()
{
  local int channel;
  local float Time;
  local float Rate;
  local Name SeqNames[16];
  local Name WasNames[16];
  local bool Changed;

  Changed = false;

  for (channel = 0; channel < 16; channel++)
    {
    Gargoyle.GetAnimParams(channel, SeqNames[channel], Time, Rate);
    if (SeqNames[channel] != PrevAnimsByChannel[channel])
      {
      Changed = true;
      WasNames[channel] = PrevAnimsByChannel[channel];
      }
    else
      WasNames[channel] = 'nc';
    }

  if (!Changed)
    return;

  log("^^^ ReportChangedAnimation ^^^ <Tick " $ TickCount $ ">" $ " Idle:" $ Gargoyle.bIsIdle);
  for (channel = 0; channel < 16; channel++)
    {
    log("    Channel " $ Channel $ ":" $ SeqNames[channel] $ " (" $ WasNames[channel] $ ")");
    PrevAnimsByChannel[channel] = SeqNames[channel];
    }
}

function Tick(float DeltaTime)
{
  local string CurrAnimPrefix;
  local string PrevAnimPrefix;
  local string CDR; // Woe has befallen those who know what this means...
  local string PrevCDR;
  local bool Managed;
  local float Time;
  local float Rate;

  TickCount++;
//log("<Tick " $ TickCount $ ">");
  if (Gargoyle == none || Gargoyle.Health <= 0)
    {
    // The gargoyle is dead and requires no more watching.
    // Commit suicide so that we don't end up with a bunch of
    // watchers scattered about the map eating up resources.
    // Unless you're godlike and never get fragged...
    Destroy();
    return;
    }

  // For some reason (probably my Y is up rig) the dual assault
  // forces the hand to rotate 90 degrees out from where it ought
  // to be.  Fix that.
// Phooey.  This merely rotates the gun, leaving the hand untouched.
  CorrectDualAssault();

//ReportChangedAnimation();
  InvokeWantedAnim();

  // Figure out what changed since the last tick, if anything.
  IsFiring = Gargoyle.Weapon.IsFiring();
  ChangedFiring = IsFiring != WasFiring;

  CurrAnim = Gargoyle.GetAnimSequence();
  CurrAnimPrefix = Left(CurrAnim, 3);
  PrevAnimPrefix = Left(PrevAnim, 3);

  Gargoyle.GetAnimParams(1, CurrAnim1, Time, Rate);
  if ((CurrAnim1 != PrevAnim1) && (CurrAnim1 == 'Weapon_Switch'))
    {
    Gargoyle.AnimBlendParams(LArmChannel, 0.0,,, LUpperArmBone);
    Gargoyle.AnimBlendParams(RArmChannel, 0.0,,, RUpperArmBone);
    }

  // Find out if this animation is one we need to deal with.
  if (
      (CurrAnimPrefix == "Run") ||
      (CurrAnimPrefix == "Wal") ||
      (CurrAnimPrefix == "Cro") ||
      (CurrAnimPrefix == "Jum") ||
      (CurrAnimPrefix == "Dou") ||
      (CurrAnimPrefix == "Idl")
     )
    Managed = true;
  else
    Managed = false;

  // For our purposes, running is defined by the animation currently
  // on channel 0.
  IsRunning = CurrAnimPrefix == "Run";
  IsCrouching = CurrAnimPrefix == "Cro";
  // This isn't exactly correct, but it's close enough for now.
  IsVertical = !IsCrouching && !IsRunning;

  ChangedCrouching = IsCrouching != WasCrouching;
  ChangedRunning = IsRunning != WasRunning;
  PrevCDR = Right(PrevAnim, 1);
  CDR = Right(CurrAnim, 1);

  ///////////////////////////////////////////////////////////
  // Clean up old animation
  ///////////////////////////////////////////////////////////
  // Clean up the firing animation if last time we were firing from a vertical pose.
  // Note that this applies only to the animations with the waist
  // twist, like walk and jump and so on.  I realized late in the
  // process that walks and jumps were designed to allow the animation
  // to be blended at the waist by making the the waist square to the
  // direction of motion.
  //
  // Note that this doesn't address situations where we're firing and transition
  // between walking and doing some other vertical thing.  One of three things
  // will happen:
  // 1. I'll change the walking animation so that the waist is twisted and it can
  //    use the same firing animation as the other vertical things.
  // 2. I'll write code that swaps between walking and non-walking vertical firing
  //    animation.
  // 3. I'll ignore the whole problem because who the hell walks in UT and will ever
  //    see the problem?
  if ((ChangedRunning && IsRunning && WasFiring) || // was firing, but running now
      (ChangedCrouching && IsCrouching && !WasRunning && WasFiring) || // was firing from a non-running stance, but crouching now
      (ChangedFiring && !IsFiring && !IsRunning && !IsCrouching))   // still vertical, but no longer firing
    StopFiringWhileVertical();

  // Clean up the firing animation if last time we were firing while running.
  else if ((ChangedRunning && !IsRunning && WasFiring) || // stopped running and was firing
           (ChangedFiring && !IsFiring && WasRunning))    // stopped firing and was running
    StopFiringWhileRunning();

  // Dunno what UT does with crouching.  Limited testing didn't tell me when it
  // uses Crouch_Burst - it used Rifle_Burst instead.
  // Which means I need to manage the frickin' crouch firing animation as well.  Nerts.
  // Clean up crouch firing if we're no longer doing both.
  else if ((ChangedCrouching && !IsCrouching && WasFiring) ||
           (ChangedFiring && !IsFiring && WasCrouching))
    StopFiringWhileCrouching();

  // There are a few scenarios under which we need to set the firing animation
  // while running:
  // - We're running and firing, and we just transitioned from a state where
  //   we weren't doing one of those things
  // - We're running and firing, and we were already, but we changed direction.
  // In these cases we need to poke UT and make it play the firing animation
  // appropriate for the direction we're headed.
  if (((ChangedRunning || ChangedFiring) && IsRunning && IsFiring) ||
      (!ChangedRunning && IsRunning && IsFiring && (CDR != PrevCDR)))
    StartFiringWhileRunning(CDR);

  // Fire up the walking firing animation if the gargoyle is currently walking
  // and firing, but wasn't doing one of those things on the previous tick.
  if ((ChangedRunning || ChangedFiring || ChangedCrouching) &&
      !IsRunning && !IsCrouching && IsFiring)
    StartFiringWhileVertical();

  if ((ChangedFiring || ChangedCrouching) && IsCrouching && IsFiring)
    StartFiringWhileCrouching();

  // Set up for next tick
  WasFiring = IsFiring;
  WasRunning = IsRunning;
  WasCrouching = IsCrouching;
}
